/*
 *
 * pgtclId.c--
 *    useful routines to convert between strings and pointers
 *  Needed because everything in tcl is a string, but we want pointers
 *  to data structures
 *
 *  ASSUMPTION:  sizeof(long) >= sizeof(void*)
 *
 *
 * Copyright (c) 1994, Regents of the University of California
 *
 *
 * IDENTIFICATION
 *    $Header: /usr/cvs/NEOSOFT/postgres95/src/libpgtcl/pgtclId.c,v 1.3 1996/08/15 19:23:56 kunkee Exp $
 *
 *-------------------------------------------------------------------------
 */

#include <stdlib.h>
#include <string.h>
#include "tcl.h"

#include "libpq-fe.h"
#include "pgtclId.h"

#define MAXHANDLES 50

typedef struct {
  struct {
	void *ptr;
	short type;
  } handle[MAXHANDLES];
  int nextHandle;
} HandleInfo;

/* convert a pointer into a string */
int
PgSetId(Tcl_Interp *interp, void *ptr, int type)
{
  HandleInfo *Pg_assoc;
  int id, current;

  Pg_assoc = (HandleInfo*)Tcl_GetAssocData(interp, PGKEY, NULL);

  current = id = Pg_assoc->nextHandle;
  while ( Pg_assoc->handle[id].ptr ) {
    if (++id == MAXHANDLES)
      id = 0;
    if (id == current) {
      interp->result = "Too many active handles";
      return TCL_ERROR;
    }
  }
  Pg_assoc->nextHandle = id;
  Pg_assoc->handle[id].ptr = ptr;
  Pg_assoc->handle[id].type = type;
  /* fprintf(stderr, "returning new id %d, ptr = %lx\n", id, (long)ptr); */
  (void) sprintf(interp->result, "pgp%d", id);
}

PgFindId(HandleInfo *Pg_assoc, void *ptr)
{
  int id;

  for (id = 0; id < MAXHANDLES; id++) {
    if (Pg_assoc->handle[id].ptr == ptr) {
      return id;
    }
  }
  return -1;
}

PgSetExistingId(Tcl_Interp *interp, void *ptr)
{
  HandleInfo *Pg_assoc;
  int id;

  Pg_assoc = (HandleInfo*)Tcl_GetAssocData(interp, PGKEY, NULL);
  id = PgFindId(Pg_assoc, ptr);
  if (id == -1) {
    (void) sprintf(interp->result, "Internal PG handle error: %lx", (long)ptr);
    return TCL_ERROR;
  } else {
      /* fprintf(stderr, "returning existing id %d, ptr = %lx\n", id, (long)ptr); */
      (void) sprintf(interp->result, "pgp%d", id);
      return TCL_OK;
  }
}


void
PgDelId(Tcl_Interp *interp, void *ptr)
{
  HandleInfo *Pg_assoc;
  int id;

  Pg_assoc = (HandleInfo*)Tcl_GetAssocData(interp, PGKEY, NULL);
  id = PgFindId(Pg_assoc, ptr);
  if (id == -1) {
      sprintf(interp->result, "Internal PG handle error: %lx", (long)ptr);
  } else {
      /* fprintf(stderr, "delete id %d, ptr = %lx\n", id, (long)ptr); */
      (void) sprintf(interp->result, "pgp%d", id);
      Pg_assoc->handle[id].ptr = 0;
  }
}


/* get back a pointer from a string */
void *
PgGetId(Tcl_Interp *interp, char *key, int type)
{
  HandleInfo *Pg_assoc;
  int id;
  void *ptr;


  if ( (strlen(key) > 3) && key[0]=='p' && key[1] == 'g' && key[2] == 'p')
  {
    Pg_assoc = (HandleInfo*)Tcl_GetAssocData(interp, PGKEY, NULL);
    id = atoi(key+3);
    if (id >= 0 && id < MAXHANDLES) {
	if (Pg_assoc->handle[id].type == type) {
          /* fprintf(stderr, "PgGetId: convert %d to %lx\n", id, Pg_assoc->handle[id].ptr); */
          return Pg_assoc->handle[id].ptr;
	}
    }
  }
  return 0;
}

/* to check to see if the string is a valid pgtcl pointer */
/* do it by attempting to get the handle and test for null */

void
Pg_cleanup(ClientData cd, Tcl_Interp *interp)
{
  HandleInfo *Pg_assoc;
  int id;
  void *ptr;


  Pg_assoc = (HandleInfo*)cd;
  for (id = MAXHANDLES - 1; id >= 0; id--) {
    if (Pg_assoc->handle[id].ptr) {
      ptr = Pg_assoc->handle[id].ptr;
      switch (Pg_assoc->handle[id].type) {
      case PG_TYPE_DBCONN:
	      /* delete connection */
	      /* fprintf(stderr, "Closing connection pgp%d, ptr %lx\n", id, ptr); */
	      PQfinish((PGconn *)ptr);
	      break;
      case PG_TYPE_RESULT:
	      /* clear result handle */
	      /* fprintf(stderr, "Clearing result pgp%d, ptr %lx\n", id, ptr); */
	      PQclear((PGresult*)ptr);
	      break;
      }
    }
  }
}

int
Pg_dump(ClientData cData, Tcl_Interp *interp, int argc, char* argv[])
{
  HandleInfo *Pg_assoc;
  int id;
  void *ptr;
  char s1[20];


  Pg_assoc = (HandleInfo*)Tcl_GetAssocData(interp, PGKEY, NULL);
  for (id = 0; id < MAXHANDLES; id++) {
    if (Pg_assoc->handle[id].ptr) {
      ptr = Pg_assoc->handle[id].ptr;
      sprintf(s1, "%d %lx %hd", id, (long)ptr, Pg_assoc->handle[id].type);
      Tcl_AppendElement(interp, s1);
    }
  }
  return TCL_OK;
}

void
Pg_InitId(Tcl_Interp *interp)
{
  HandleInfo *Pg_assoc;
  int i;

  if (!Tcl_GetAssocData(interp, PGKEY, NULL))
  {
	Pg_assoc = (HandleInfo*)ckalloc(sizeof(HandleInfo));
	for (i = 0; i < MAXHANDLES; i++) {
		Pg_assoc->handle[i].ptr = 0;
		Pg_assoc->handle[i].type = 0;
	}
	Pg_assoc->nextHandle = 0;
	Tcl_SetAssocData(interp, PGKEY, Pg_cleanup, Pg_assoc);
  }
}

/* Pass the PG connection info from the trusted interp to the safe one */
int
Pg_handle(ClientData cData, Tcl_Interp *interp, int argc, char **argv)
{
  Tcl_Interp *slaveInterp;
  HandleInfo *Pg_assoc;
  void *ptr;
  char *key;
  int id;

  if (argv[1] == NULL || argv[2] == NULL) {
	sprintf(interp->result, "Usage: Pg_handle interp handle_name");
	return TCL_ERROR;
  }

  Pg_assoc = (HandleInfo*)Tcl_GetAssocData(interp, PGKEY, NULL);

  key = argv[2];
  if ( (strlen(key) > 3) && key[0]=='p' && key[1] == 'g' && key[2] == 'p')
  {
    id = atoi(key+3);
    if (id < 0 && id >= MAXHANDLES) {
        return TCL_ERROR;
    }
  }

  slaveInterp = Tcl_GetSlave(interp, argv[1]);
  PgSetId(slaveInterp, Pg_assoc->handle[id].ptr, Pg_assoc->handle[id].type);
  PgDelId(interp, Pg_assoc->handle[id].ptr);
  return TCL_OK;
}
